/* ChanUser.c - NeonServ v5.5
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#include "ChanUser.h"
#include "ChanNode.h"
#include "ModeNode.h"
#include "UserNode.h"

struct ChanUser* addChanUser(struct ChanNode *chan, struct UserNode *user) {
    struct ChanUser *chanuser = malloc(sizeof(*chanuser));
    if (!chanuser)
    {
        perror("malloc() failed");
        return NULL;
    }
    chanuser->flags = 0;
    chanuser->user = user;
    chanuser->chan = chan;
    chanuser->visCount = 0;
    
    chanuser->changeTime = 0;
    chanuser->spamnode = NULL;

    SYNCHRONIZE(cache_sync);

    chanuser->next_user = chan->user;
    chan->user = chanuser;
    chan->usercount++;

    chanuser->next_chan = user->channel;
    user->channel = chanuser;

    DESYNCHRONIZE(cache_sync);

    return chanuser;
}

struct ChanUser* addInvisibleChanUser(struct ChanNode *chan, struct UserNode *user) {
    struct ChanUser *chanuser = malloc(sizeof(*chanuser));
    if (!chanuser)
    {
        perror("malloc() failed");
        return NULL;
    }
    chanuser->flags = CHANUSERFLAG_INVISIBLE;
    chanuser->user = user;
    chanuser->chan = chan;
    chanuser->visCount = 0;
    
    chanuser->changeTime = 0;
    chanuser->spamnode = NULL;

    SYNCHRONIZE(cache_sync);
    chanuser->next_user = chan->user;
    chan->user = chanuser;
    DESYNCHRONIZE(cache_sync);
    chan->usercount++;

    return chanuser;
}

int isUserOnChan(struct UserNode *user, struct ChanNode *chan) {
    struct ChanUser *chanuser;
    if(isModeSet(chan->modes, 'd') || isModeSet(chan->modes, 'D')) {
        for(chanuser = chan->user; chanuser; chanuser = chanuser->next_user) {
            if(chanuser->user == user)
                return 1;
        }
    } else {
        for(chanuser = user->channel; chanuser; chanuser = chanuser->next_chan) {
            if(chanuser->chan == chan)
                return 1;
        }
    }
    return 0;
}

struct ChanUser* getChanUser(struct UserNode *user, struct ChanNode *chan) {
    struct ChanUser *chanuser;
    if(isModeSet(chan->modes, 'd') || isModeSet(chan->modes, 'D')) {
        for(chanuser = chan->user; chanuser; chanuser = chanuser->next_user) {
            if(chanuser->user == user)
                return chanuser;
        }
    } else {
        for(chanuser = user->channel; chanuser; chanuser = chanuser->next_chan) {
            if(chanuser->chan == chan)
                return chanuser;
        }
    }
    return NULL;
}

struct ChanUser* getChannelUsers(struct ChanNode *chan, struct ChanUser *last) {
    if(last == NULL)
        return chan->user;
    else
        return last->next_user;
}

struct ChanUser* getUserChannels(struct UserNode *user, struct ChanUser *last) {
    if(last == NULL)
        return user->channel;
    else
        return last->next_chan;
}

void delChanUser(struct ChanUser *chanuser, int do_freeChanUser) {
    SYNCHRONIZE(cache_sync);
    struct ChanUser *cchanuser, *last;
    //remove it from the user's channel-list
    last = NULL;
    for(cchanuser = chanuser->user->channel; cchanuser; cchanuser = cchanuser->next_chan) {
        if(cchanuser == chanuser) {
            if(last) 
                last->next_chan = chanuser->next_chan;
            else
                chanuser->user->channel = chanuser->next_chan;
            break;
        } else
            last = cchanuser;
    }

    //remove it from the channel's user-list
    last = NULL;
    for(cchanuser = chanuser->chan->user; cchanuser; cchanuser = cchanuser->next_user) {
        if(cchanuser == chanuser) {
            chanuser->chan->usercount--;
            if(last) 
                last->next_user = chanuser->next_user;
            else
                chanuser->chan->user = chanuser->next_user;
            break;
        } else
            last = cchanuser;
    }
    
    if(do_freeChanUser) {
        freeChanUser(chanuser);
    } else {
        chanuser->next_chan = NULL;
        chanuser->next_user = NULL;
    }
    DESYNCHRONIZE(cache_sync);
}

void removeChanUserFromLists(struct ChanUser *chanuser, int remove_from_userlist, int remove_from_channellist, int do_freeChanUser) {
    SYNCHRONIZE(cache_sync);
    struct ChanUser *cchanuser, *last;
    if(remove_from_userlist) {
        //remove it from the channel's user-list
        last = NULL;
        for(cchanuser = chanuser->chan->user; cchanuser; cchanuser = cchanuser->next_user) {
            if(cchanuser == chanuser) {
                chanuser->chan->usercount--;
                if(last) 
                    last->next_user = chanuser->next_user;
                else
                    chanuser->chan->user = chanuser->next_user;
                break;
            } else
                last = cchanuser;
        }
        chanuser->next_user = NULL;
    }
    if(remove_from_channellist) {
        //remove it from the user's channel-list
        last = NULL;
        for(cchanuser = chanuser->user->channel; cchanuser; cchanuser = cchanuser->next_chan) {
            if(cchanuser == chanuser) {
                if(last) 
                    last->next_chan = chanuser->next_chan;
                else
                    chanuser->user->channel = chanuser->next_chan;
                break;
            } else
                last = cchanuser;
        }
        chanuser->next_chan = NULL;
    }
    
    if(do_freeChanUser) {
        freeChanUser(chanuser);
    }
    DESYNCHRONIZE(cache_sync);
}

void freeChanUser(struct ChanUser *chanuser) {
    if(chanuser->spamnode)
        free(chanuser->spamnode);
    free(chanuser);
}

